import os
import re
from pathlib import Path


class Formatter(object):
    def __init__(self,
                 content,
                 style=None,
                 output=None,
                 py_prompt='shell',
                 newline_between_headers=False,
                 reindex_headers=True):
        self.content = content
        self.style = style
        self.output = output
        self.py_prompt = py_prompt
        self.newline_between_headers = newline_between_headers
        self.reindex_headers = reindex_headers
        self.links = {}
        self.new_lines = []
        self.table_started = False
        self.code_block_started = False
        self.code_quote = '`'
        self.header_formatter = HeaderFormatter(self.new_lines)

        if self.style != 'code':
            self.code_quote = ''

        if self.py_prompt == 'shell':
            self.set_py_prompt = self.set_py_prompt_shell
        elif self.py_prompt == 'ipython':
            self.set_py_prompt = self.set_py_prompt_ipython
        else:
            raise ValueError('unknown py_prompt')

    def format(self):
        if self.new_lines:
            return self.output_result()

        if hasattr(self.content, 'read'):
            self.content = self.content.read()

        raw_lines = self.content.splitlines()

        for index, line in enumerate(raw_lines):

            if self.is_in_code_block(line):
                line = self.set_py_prompt(line)
                self.new_lines.append(line)
                continue

            if line.isspace() or not line:
                if self.new_lines and self.new_lines[-1].startswith('>'):
                    self.new_lines.append('\n')
                continue

            if self.is_split(line):
                self.new_lines.append(line)
                continue

            is_header = self.is_header(line)

            if is_header and self.new_lines and not self.is_header(self.new_lines[-1]):
                if self.newline_between_headers:
                    self.new_lines.append('\n<br/>\n')

            if '|' in line:
                if not self.table_started:
                    try:
                        next_line = raw_lines[index + 1]
                    except IndexError:
                        next_line = ''

                    if re.match(r'^-+?|-+?', next_line):
                        self.table_started = True

            elif self.table_started:
                self.table_started = False

            if is_header:
                self.header_formatter.add_header()

            if self.table_started:
                self.new_lines.append(line)
            else:
                self.new_lines.append(self.format_line(line))

        if self.reindex_headers:
            self.header_formatter.set()

        return self.output_result()

    def output_result(self):
        new_text = '\n'.join(self.new_lines)

        if self.output is None:
            return new_text

        if self.output == 'stream':
            return print(new_text)

        with Path(self.output).open('w', encoding='utf-8') as f:
            f.write(new_text)

    def is_header(self, line):
        return re.match(r'^#+?\s', line)

    def is_split(self, line):
        """判断该行是否为分割符"""
        return line.startswith('--') and '|' not in line

    def is_in_code_block(self, line):
        """判断是否处在代码块中"""
        if not line.strip().startswith('```'):
            return self.code_block_started

        if not self.code_block_started:
            self.code_block_started = True
        else:
            self.code_block_started = False

        return True

    def set_py_prompt_shell(self, line):
        """将 python 的提示符设置为 shell 的风格
        ex: ">>> print(\n)"
        """
        line = re.sub(r'^In \[\d+\]: ', '>>> ', line)
        line = re.sub(r'^Out\[\d+\]: ', '', line)
        line = re.sub(r'^\s*\.\.\.: ', '... ', line)
        return line

    def set_py_prompt_ipython(self, line):
        """将 python 的提示符设置为 ipython 的风格
        ex:
            [in 1] print(\n)
            [out1]

        """
        return line

    def remove_links(self, line):
        """移除内容中的链接"""
        self.links = {}

        links = re.findall(r'!?\[.*?\]\(.*?\)', line)

        for index, link in enumerate(set(links)):
            placeholder = ' <LinkStashed(num=%s)> ' % index
            line = line.replace(link, placeholder)
            self.links[link] = placeholder

        return line

    def recover_links(self, line):
        for link, placeholder in self.links.items():
            line = line.replace(placeholder, link)

        self.links = {}

        return line

    def is_zh(self, string):
        """判断字符是否为中文字符"""
        return string and ('\u4e00' <= string <= '\u9fa5' or self.is_zh_mark(string))

    def is_zh_mark(self, string):
        """判断字符是否为中文标点符号"""
        return string in '，。（）：、'

    def is_mark(self, string):
        return string and (self.is_zh_mark(string) or self.is_en_mark(string))

    def is_en(self, string):
        return re.match(r'[a-zA-Z0-9_\\\[\]\{\}\*\'&#\$\(\)]', string)

    def is_en_mark(self, string):
        return string in ("'", '"', ':', ';', '(', ')', '/')

    def format_line(self, line):
        line = self.remove_links(line)
        new_words = []
        half_quoted = False

        def next_non_blank_word(idx):
            """获取下一个非空白字符和索引"""
            next_index = idx + 1
            next_word = ''
            while next_index < len(line):
                next_word = line[next_index]
                if not next_word.isspace():
                    break
                next_index += 1
            return next_word, next_index

        def prev_non_blank_word(idx):
            """获取前一个非空白字符和索引"""
            pre_index = idx - 1
            pre_word = ''
            while pre_index >= 0:
                pre_word = line[pre_index]
                if not pre_word.isspace():
                    break
                pre_index -= 1
            return pre_word, pre_index

        def is_bold_quote(word, index):
            if word != '*':
                return False

            next_word, next_index = next_non_blank_word(index)
            return next_word == word and next_index == index + 1

        for index, word in enumerate(line):

            if not self.require_space(word):
                new_words.append(word)
                continue

            def add_quote():
                if word != '`':
                    new_words.append(self.code_quote)

            if not index or all(w.isspace() for w in new_words):
                # 第一个非空白字符
                if word in '#-' or word.isdigit():
                    new_words.append(word)
                elif is_bold_quote(word, index):
                    new_words.append(word)
                    continue
                else:
                    add_quote()
                    new_words.append(word)
                    half_quoted = True

            elif half_quoted:
                # 下一个非空白字符
                next_word, next_index = next_non_blank_word(index)
                blank_exists = next_index > index + 1

                if not next_word:
                    new_words.append(word)
                    add_quote()
                    half_quoted = False
                    continue

                if self.require_space(next_word):
                    new_words.append(word)

                else:
                    new_words.append(word)
                    add_quote()
                    if not blank_exists and not self.is_zh_mark(next_word):
                        new_words.append(' ')
                    half_quoted = False

            elif is_bold_quote(word, index):
                new_words.append(word)

            else:
                # 根据前后字符处理
                # 前一个字符
                # pre_word = line[index - 1]
                prev_word, prev_index = prev_non_blank_word(index)
                prev_blank_exists = prev_index + 1 < index

                # 下一个非空白字符
                next_word, next_index = next_non_blank_word(index)
                blank_exists = next_index > index + 1

                if self.is_zh(prev_word):
                    if self.is_en_mark(word) and (not next_word or self.is_zh(next_word)):
                        new_words.append(word)
                        continue

                    if not self.is_zh_mark(prev_word) and not prev_blank_exists:
                        new_words.append(' ')

                    add_quote()
                    new_words.append(word)

                    if self.is_zh(next_word):
                        add_quote()
                        if not self.is_zh_mark(next_word) and not blank_exists:
                            new_words.append(' ')
                    else:
                        half_quoted = True

                    continue

                # 如果前面字符不需要处理，看后面的字符
                if self.is_zh(next_word) and not self.is_zh_mark(next_word):
                    new_words.append(word)
                    if not blank_exists:
                        new_words.append(' ')
                    continue

                new_words.append(word)

        line = ''.join(new_words)

        return self.recover_links(line)

    def require_space(self, string):
        """判断字符是否为需要添加空白的字符"""
        return self.is_en(string) or self.is_en_mark(string) or string in ['-', '`', '=', '.', ',']


class HeaderFormatter(object):

    def __init__(self, lines):
        self.lines = lines
        self.headers = []
        self.current_index = []
        self.max_level = None

    def add_header(self, index=None):
        if index is None:
            index = len(self.lines)
        self.headers.append(index)

    def set(self):
        if not self.headers:
            return

        self.current_index = [0]

        levels = {}

        for index in self.headers:
            level = self.get_level(index)
            levels[index] = level

        self.top_level = min(level for level in levels.values())

        for index, level in levels.items():
            header_index = self.get_index(level)
            self.set_index(index, header_index)

    def get_level(self, index):
        return len(self.lines[index].split(maxsplit=1)[0])

    def get_index(self, level):
        missing = level - self.top_level
        if not missing:
            self.current_index = [self.current_index[0] + 1]
            return self.make_index(self.current_index[0])

        require_len = missing + 1

        if len(self.current_index) < require_len:
            for _ in range(require_len - len(self.current_index)):
                self.current_index.append(0)

        self.current_index[missing] += 1
        return self.make_index(self.current_index[:require_len])

    def make_index(self, index):
        if isinstance(index, int):
            return str(index)

        return '.'.join(str(i) for i in index)

    def set_index(self, line_index, header_index):
        line = self.lines[line_index]
        prefix, line = line.split(maxsplit=1)
        line = self.remove_exists_index(line)
        line = f'{prefix} {header_index}. {line}'
        self.lines[line_index] = line

    def remove_exists_index(self, line):
        if re.match(r'^\d[\d\.]+\.\s', line):
            return line.split(maxsplit=1)[-1]
        return line


def format(*args, **kwargs):
    return Formatter(*args, **kwargs).format()


def format_file(filepath, **kwargs):
    _, name = os.path.split(filepath)
    kwargs.setdefault('output', name)

    with open(filepath, 'r', encoding='utf-8') as f:
        return format(f, **kwargs)



from textwrap import dedent
from unittest import TestCase

from prettymd.formatter import format


class TestFormatter(TestCase):

    def assert_formatted(self, text, expect, **kwargs):
        kwargs.setdefault('style', 'code')
        self.assertEqual(expect, format(text, output=None, **kwargs))

    def test_en_with_mark_formatted(self):
        text = '默认的表名是appName modelName，在'
        expect = '默认的表名是 `appName modelName`，在'
        self.assert_formatted(text, expect)

    def test_code_block_beautify(self):
        text = """\t```python
            def fn():

                pass
            \t```"""

        self.assert_formatted(text, text)

    def test_quote_en_words(self):
        text = '摘要算法就是通过摘要函数f()对任意长度的数据data计算出固定长度的摘要digest，目的是为了发现原始数据是否被人篡改过。'
        expect = '摘要算法就是通过摘要函数 `f()` 对任意长度的数据 `data` 计算出固定长度的摘要 `digest`，目的是为了发现原始数据是否被人篡改过。'

        self.assert_formatted(text, expect)

    def test_en_words_on_start(self):
        text = 'SHA1的结果是160 bit字节'
        expect = '`SHA1` 的结果是 `160 bit` 字节'
        self.assert_formatted(text, expect)

        text = '# 1'
        self.assert_formatted(text, text, reindex_headers=False)

    def test_table_should_not_be_beautified(self):
        text = """
        name    | password
        --------|----------
        michael | 123456
        """

        text = dedent(text).strip()
        self.assert_formatted(text, text)

    def test_link_be_ignored(self):
        text = """
        1. [codecs --- 编解码器注册和相关基类 — Python 3.12.4 文档](https://docs.python.org/zh-cn/3/library/codecs.html)
        """
        text = dedent(text).strip()
        self.assert_formatted(text, text)

    def test_word_after_number_will_be_formatted(self):
        text = '1. echo将字符串"dev"通过管道符传递给标准输出，-n选项可以去掉多余的换行符'
        expect = '1. echo 将字符串 `"dev"` 通过管道符传递给标准输出，`-n` 选项可以去掉多余的换行符'

        self.assert_formatted(text, expect)

    def test_en_word_in_zh_mark(self):
        text = "由于结果中同时也包含了文件名（在本例中文件名为-），通过指定分割符-d' '将输出拆分为两列"
        expect = "由于结果中同时也包含了文件名（在本例中文件名为 `-`），通过指定分割符 `-d' '` 将输出拆分为两列"
        self.assert_formatted(text, expect)

        text = "可以使用md5sum、sha224sum、sha1sum 等"
        expect = "可以使用 `md5sum`、`sha224sum`、`sha1sum` 等"
        self.assert_formatted(text, expect)

    def test_quote_has_blank(self):
        text = "在常规`models`操作中，某些对于关联表的查询稍有不慎可能导致产生多次查询，如:"
        expect = "在常规 `models` 操作中，某些对于关联表的查询稍有不慎可能导致产生多次查询，如:"
        self.assert_formatted(text, expect)

    def test_code_has_expr_in_quotes(self):
        text = "上面这种查询操作会造成不必要的多次查询，因为当`userlist = models.User.objects.all()`执行时"
        expect = "上面这种查询操作会造成不必要的多次查询，因为当 `userlist = models.User.objects.all()` 执行时"
        self.assert_formatted(text, expect)

    def test_parenthesis_are_ignored(self):
        text = "次查询(即JOIN查询)，"
        expect = "次查询(即 `JOIN` 查询)，"
        self.assert_formatted(text, expect)

    def test_multiple_en_formatted(self):
        text = 'MASTER 和 BINARY 是同义词'
        expect = '`MASTER` 和 `BINARY` 是同义词'
        self.assert_formatted(text, expect)

    def test_endswith_en_formatted(self):
        text = '#### 关闭 mysql 主从，关闭 binlog'
        expect = '#### 关闭 `mysql` 主从，关闭 `binlog`'
        self.assert_formatted(text, expect, reindex_headers=False)

    def test_no_br_in_headers(self):
        text = """
        ## h2
        ### h3
        """

        expect = """
        ## h2
        ### h3
        """

        self.assert_formatted(dedent(text), dedent(expect).strip(), reindex_headers=False)

    def test_br_between_in_headers(self):
        text = """
        你好
        ## h2
        ### h3
        """

        expect = """
        你好

        <br/>

        ## h2
        ### h3
        """

        self.assert_formatted(dedent(text), dedent(expect).strip(), reindex_headers=False, newline_between_headers=True)

    def test_code_with_slash(self):
        text = "使用 `git reset --hard`或 `git checkout -t -f remote/branch`"
        expect = "使用 `git reset --hard` 或 `git checkout -t -f remote/branch`"
        self.assert_formatted(text, expect)

    def test_bold_text(self):
        text = "**1. `Django`的`QuerySet`是惰性的。**`Django`的`QuerySet`对应"
        expect = "**1. `Django` 的 `QuerySet` 是惰性的。**`Django` 的 `QuerySet` 对应"
        self.assert_formatted(text, expect)

    def test_parenthesis_is_quoted(self):
        text = "默认值为 (80, 24)，如果"
        expect = "默认值为 `(80, 24)`，如果"
        self.assert_formatted(text, expect)

    def test_title_reformat(self):
        text = """
        ## h2
        ### h3
        ## h22
        #### h4
        ### h23
        """

        expect = """
        ## 1. h2
        ### 1.1. h3
        ## 2. h22
        #### 2.0.1. h4
        ### 2.1. h23
        """

        self.assert_formatted(dedent(text), dedent(expect).strip(), reindex_headers=True)



import sys

from argparse import ArgumentParser
from prettymd import format, format_file


def main():
    parser = ArgumentParser()
    parser.add_argument('args', metavar='ARG', type=str, nargs='*')
    parser.add_argument('-f --file', dest='file', type=str)
    parser.add_argument('-o --output', dest='output', default='stream')
    parser.add_argument('-r --reindex-headers', dest='reindex_headers', action='store_true', default=False)
    parser.add_argument('-p --py-prompt', dest='py_prompt', default='shell')
    parser.add_argument('-n --newline-between-headers', dest='newline_between_headers', action='store_true', default=False)
    parser.add_argument('-s --style', dest='style', default=None)

    args = parser.parse_args(args=sys.argv[1:])
    kwargs = dict(output=args.output,
                  style=args.style,
                  reindex_headers=args.reindex_headers,
                  py_prompt=args.py_prompt,
                  newline_between_headers=args.newline_between_headers)

    if args.file:
        format_file(args.file, **kwargs)
    elif args.args:
        for content in args.args:
            format(content, **kwargs)
    else:
        print('No content specified.')
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

from os import path

base_dir = path.abspath(path.dirname(__file__))

with open(path.join(base_dir, 'README.md'), encoding='utf8') as readme_file:
    readme = readme_file.read()

with open(path.join(base_dir, 'HISTORY.md'), encoding='utf8') as history_file:
    history = history_file.read()

requirements = []

setup_requirements = []

test_requirements = []

setup(
    author="kingron",
    author_email='kingronjan@qq.com',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="A tool that formats markdown text in Chinese.",
    install_requires=requirements,
    long_description=readme + '\n\n' + history,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='prettymd',
    name='prettymd',
    packages=find_packages(include=['prettymd', 'prettymd.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/kingronjan/prettymd',
    version='0.1.11',
    zip_safe=False,
)
